import CELVisitor from './generated/CELVisitor';
import { builtInFunctions } from './BuiltInFunctions';
import Context from './Context';
/**
 * Validates if a string represents a valid timestamp in ISO 8601 format.
 * Accepts both date-only (YYYY-MM-DD) and full datetime formats.
 */
export function isValidTimestampString(str) {
    // Accepts ISO 8601 date or datetime strings
    const parsed = Date.parse(str);
    if (isNaN(parsed))
        return false;
    // Check if the parsed date's ISO string starts with the input (date or datetime)
    const d = new Date(parsed);
    const iso = d.toISOString();
    // Accepts date-only (YYYY-MM-DD) or full ISO
    return (str.length === 10 && iso.startsWith(str) ||
        iso.startsWith(str));
}
export class Evaluator extends CELVisitor {
    constructor(context = new Context({})) {
        super();
        this.visitStart = (ctx) => {
            return this.visit(ctx.expr());
        };
        this.visitExpr = (ctx) => {
            return this.visit(ctx.getChild(0));
        };
        this.visitConditionalOr = (ctx) => {
            if (ctx.children.length === 1) {
                return this.visit(ctx.getChild(0));
            }
            else if (ctx.children.length === 5 && ctx.getChild(1).getText() === '?') {
                const condition = this.visit(ctx.getChild(0));
                return condition ? this.visit(ctx.getChild(2)) : this.visit(ctx.getChild(4));
            }
            else {
                let result = this.visit(ctx.getChild(0));
                for (let i = 1; i < ctx.children.length; i += 2) {
                    const right = this.visit(ctx.getChild(i + 1));
                    result = result || right;
                    if (result)
                        break;
                }
                return result;
            }
        };
        this.visitConditionalAnd = (ctx) => {
            if (ctx.children.length === 1) {
                return this.visit(ctx.getChild(0));
            }
            else {
                let result = this.visit(ctx.getChild(0));
                for (let i = 1; i < ctx.children.length; i += 2) {
                    const right = this.visit(ctx.getChild(i + 1));
                    result = result && right;
                    if (!result)
                        break;
                }
                return result;
            }
        };
        this.visitRelationOp = (ctx) => {
            const left = this.visit(ctx.getChild(0)); // Evaluate the left operand
            const operator = ctx.getChild(1).getText(); // Get the operator (e.g., '<=')
            const right = this.visit(ctx.getChild(2)); // Evaluate the right operand
            return this.evaluateComparison(operator, left, right);
        };
        this.visitRelationCalc = (ctx) => {
            return this.visit(ctx.getChild(0));
        };
        this.visitCalcAddSub = (ctx) => {
            if (ctx.children.length === 1) {
                return this.visit(ctx.getChild(0));
            }
            else {
                let result = this.visit(ctx.getChild(0));
                for (let i = 1; i < ctx.children.length; i += 2) {
                    const operator = ctx.getChild(i).getText();
                    const right = this.visit(ctx.getChild(i + 1));
                    switch (operator) {
                        case '+':
                            result = result + right;
                            break;
                        case '-':
                            result = result - right;
                            break;
                        default:
                            throw new Error(`Unknown operator: ${operator}`);
                    }
                }
                return result;
            }
        };
        this.visitCalcMulDiv = (ctx) => {
            if (ctx.children.length === 1) {
                return this.visit(ctx.getChild(0));
            }
            else {
                let result = this.visit(ctx.getChild(0));
                for (let i = 1; i < ctx.children.length; i += 2) {
                    const operator = ctx.getChild(i).getText();
                    const right = this.visit(ctx.getChild(i + 1));
                    switch (operator) {
                        case '*':
                            result = result * right;
                            break;
                        case '/':
                            result = result / right;
                            break;
                        case '%':
                            result = result % right;
                            break;
                        default:
                            throw new Error(`Unknown operator: ${operator}`);
                    }
                }
                return result;
            }
        };
        this.visitCalcUnary = (ctx) => {
            if (ctx.children.length === 1) {
                return this.visit(ctx.getChild(0));
            }
            else {
                const operator = ctx.getChild(0).getText();
                const operand = this.visit(ctx.getChild(1));
                switch (operator) {
                    case '!':
                        return !operand;
                    case '-':
                        return -operand;
                    default:
                        throw new Error(`Unknown unary operator: ${operator}`);
                }
            }
        };
        this.visitLogicalNot = (ctx) => {
            const value = this.visit(ctx.getChild(1));
            return !value;
        };
        this.visitNegate = (ctx) => {
            const value = this.visit(ctx.getChild(1));
            return -value;
        };
        this.visitMemberExpr = (ctx) => {
            return this.visit(ctx.getChild(0));
        };
        this.visitSelectOrCall = (ctx) => {
            if (ctx.children.length === 1) {
                return this.visit(ctx.getChild(0));
            }
            const object = this.visit(ctx.getChild(0));
            const member = ctx.getChild(2).getText();
            if (ctx.children.length === 3) {
                if (object && typeof object === 'object' && member in object) {
                    return object[member];
                }
                else {
                    throw new Error(`Property '${member}' does not exist on object`);
                }
            }
            else if (ctx.children.length >= 4 && ctx.getChild(3).getText() === '(') {
                let args = [];
                if (ctx.children.length === 5) {
                    args = this.visitExprList(ctx.getChild(4));
                }
                if (builtInFunctions.hasOwnProperty(member)) {
                    try {
                        // @ts-ignore              
                        return builtInFunctions[member](object, args);
                    }
                    catch (error) {
                        throw new Error(`Error in method '${member}': ${error.message}`);
                    }
                }
                else {
                    throw new Error(`Method '${member}' is not defined`);
                }
            }
            else {
                throw new Error('Invalid member expression');
            }
        };
        this.visitPrimaryExpr = (ctx) => {
            if (ctx.children.length === 1) {
                return this.visit(ctx.getChild(0));
            }
            else if (ctx.children.length === 3 && ctx.getChild(0).getText() === '(') {
                return this.visit(ctx.getChild(1));
            }
            else {
                throw new Error('Invalid primary expression');
            }
        };
        this.visitIdentOrGlobalCall = (ctx) => {
            const id = ctx.getChild(0).getText();
            if (ctx.children.length === 1) {
                const variableValue = this.context.getVariable(id);
                if (variableValue !== undefined) {
                    return variableValue;
                }
                throw new Error(`Variable '${id}' is not defined`);
            }
            else if (ctx.children.length >= 3 && ctx.getChild(1).getText() === '(') {
                let args = [];
                if (ctx.children.length === 4) {
                    args = this.visitExprList(ctx.getChild(2));
                }
                if (builtInFunctions.hasOwnProperty(id)) {
                    try {
                        // @ts-ignore  
                        return builtInFunctions[id](...args);
                    }
                    catch (error) {
                        throw new Error(`Error in function '${id}': ${error.message}`);
                    }
                }
                else {
                    throw new Error(`Function '${id}' is not defined`);
                }
            }
            else {
                throw new Error('Invalid function call');
            }
        };
        this.visitNested = (ctx) => {
            return this.visit(ctx.expr());
        };
        this.visitExprList = (ctx) => {
            const exprs = [];
            for (let i = 0; i < ctx.children.length; i += 2) {
                exprs.push(this.visit(ctx.getChild(i)));
            }
            return exprs;
        };
        this.visitConstantLiteral = (ctx) => {
            return this.visit(ctx.getChild(0));
        };
        this.visitInt = (ctx) => {
            const text = ctx.getText();
            return parseInt(text, 10);
        };
        this.visitUint = (ctx) => {
            const text = ctx.getText();
            return parseInt(text, 10);
        };
        this.visitDouble = (ctx) => {
            const text = ctx.getText();
            return parseFloat(text);
        };
        this.visitString = (ctx) => {
            let text = ctx.getText();
            text = text.substring(1, text.length - 1).replace(/\\'/g, "'");
            return text;
        };
        this.visitBoolTrue = (ctx) => {
            return true;
        };
        this.visitBoolFalse = (ctx) => {
            return false;
        };
        this.visitNull = (ctx) => {
            return null;
        };
        this.visitIdent = (ctx) => {
            const id = ctx.getText();
            const variableValue = this.context.getVariable(id);
            if (variableValue !== undefined) {
                return variableValue;
            }
            throw new Error(`Variable '${id}' is not defined`);
        };
        this.visitConditionalExpr = (ctx) => {
            if (ctx.children.length === 1) {
                return this.visit(ctx.getChild(0));
            }
            else {
                const condition = this.visit(ctx.getChild(0));
                return condition ? this.visit(ctx.getChild(2)) : this.visit(ctx.getChild(4));
            }
        };
        this.context = context;
    }
    setContext(context) {
        this.context = context;
    }
    evaluateComparison(operator, left, right) {
        // Only transform timestamp strings directly to numeric timestamps
        if (typeof left === 'string' && isValidTimestampString(left)) {
            left = new Date(left).getTime();
        }
        if (typeof right === 'string' && isValidTimestampString(right)) {
            right = new Date(right).getTime();
        }
        switch (operator) {
            case '==':
                return left === right;
            case '!=':
                return left !== right;
            case '<=':
                return left <= right;
            case '<':
                return left < right;
            case '>=':
                return left >= right;
            case '>':
                return left > right;
            default:
                throw new Error(`Unsupported operator: ${operator}`);
        }
    }
}
export default Evaluator;
