import antlr4 from 'antlr4';
import CELLexer from './generated/CELLexer.js';
import CELParser from './generated/CELParser.js';
import Evaluator from './Evaluator';
import Context from './Context';
import { ErrorCollector } from './ErrorCollector';
import TypeChecker from './TypeChecker';
export class Runtime {
    constructor(celExpression) {
        this.ast = null;
        this.errors = [];
        // const chars = new antlr4.InputStream(celExpression);
        // const lexer = new CELLexer(chars);
        this.celExpression = celExpression;
        const chars = antlr4.CharStreams.fromString(celExpression);
        const lexer = new CELLexer(chars);
        const tokens = new antlr4.CommonTokenStream(lexer);
        const parser = new CELParser(tokens);
        parser.buildParseTrees = true;
        parser.removeErrorListeners();
        const errorCollector = new ErrorCollector();
        parser.addErrorListener(errorCollector);
        try {
            this.ast = parser.start();
        }
        catch (error) {
            this.ast = null;
        }
        if (errorCollector.errors.length > 0) {
            this.errors = errorCollector.errors;
            this.ast = null;
        }
    }
    static canParse(celExpression) {
        const runtime = new Runtime(celExpression);
        return runtime.ast !== null;
    }
    static parseString(celExpression) {
        const runtime = new Runtime(celExpression);
        if (runtime.ast !== null) {
            return { success: true };
        }
        else {
            return {
                success: false,
                error: runtime.errors[0]?.message || 'Parsing failed with errors',
            };
        }
    }
    static typeCheck(expression, context, types) {
        const runtime = new Runtime(expression);
        if (runtime.ast !== null) {
            try {
                let contextObj;
                if (context instanceof Context) {
                    contextObj = context;
                }
                else {
                    contextObj = new Context(context || {}, types || {});
                }
                const typeChecker = new TypeChecker(contextObj);
                typeChecker.visit(runtime.ast);
                return { success: true };
            }
            catch (error) {
                return { success: false, error: error.message };
            }
        }
        else {
            return { success: false, error: 'Parsing failed with errors' };
        }
    }
    evaluate(context, types) {
        if (!this.ast) {
            throw new Error('AST is not available. Parsing might have failed.');
        }
        const contextObj = new Context(context, types);
        const typeCheckResult = Runtime.typeCheck(this.celExpression, contextObj);
        if (typeCheckResult.success) {
            const evaluator = new Evaluator(contextObj);
            return evaluator.visit(this.ast);
        }
        else {
            throw new Error(typeCheckResult.error);
        }
    }
}
